import os
import shutil

from dbt_rpc.contracts.rpc import (
    RPCDepsParameters, RemoteDepsResult, RemoteMethodFlags,
)
from dbt_rpc.rpc.method import RemoteMethod
from dbt.task.deps import DepsTask


def _clean_deps(config):
    if os.path.exists(config.packages_install_path):
        shutil.rmtree(config.packages_install_path)
    os.makedirs(config.packages_install_path)


class RemoteDepsTask(
    RemoteMethod[RPCDepsParameters, RemoteDepsResult],
    DepsTask,
):
    METHOD_NAME = 'deps'

    def get_flags(self) -> RemoteMethodFlags:
        return (
            RemoteMethodFlags.RequiresConfigReloadBefore |
            RemoteMethodFlags.RequiresManifestReloadAfter
        )

    def set_args(self, params: RPCDepsParameters):
        pass

    def handle_request(self) -> RemoteDepsResult:
        _clean_deps(self.config)
        self.run()
        return RemoteDepsResult([])



"""The `rpc` package handles most aspects of the actual execution of dbt's RPC
server (except for the server itself and the client tasks, which are defined in
the `task.remote` package).

The general idea from a thread/process management perspective (ignoring the
--single-threaded flag!) is as follows:

- The RPC server runs a web server, in particular `werkzeug`, which manages a
  thread pool.
- When a request comes in, werkzeug spins off a thread to manage the
  request/response portion. dbt itself has basically no control over this
  operation - from our viewpoint request/response cycles are fully
  synchronous.
- synchronous requests are defined as methods in the `TaskManager` and handled
  in the responding thread directly.
- Asynchronous requests (defined in `tasks.remote`) are kicked off wrapped in
  `RequestTaskHandler`s, which manage a new process and a new thread.
    - The process runs the actual dbt request, logging via a message queue
    - eventually just before process exit, the process places an "error" or
      "result" on the queue
    - The thread monitors the queue, taking logs off the queue and adding them
      to the `RequestTaskHandler`'s `logs` attribute.
    - The thread also monitors the `is_alive` state of the process, in case
      it is killed "unexpectedly" (including via `kill`)
    - When the thread sees an error or result come over the queue, it join()s
      the process.
    - When the thread sees that the process has disappeared without placing
      anything on the queue, it checks the queue one last time, and then acts
      as if the queue received an 'Unexpected termination' error
- `kill` commands pointed at an asynchronous task kill the process and allow
  the thread to handle cleanup and management
- When the RPC server receives a shutdown instruction, it:
  - stops responding to requests
  - `kills` all processes (triggering the end of all processes, right!?)
  - exits (all remaining threads should die here!)
"""



from abc import abstractmethod
from datetime import datetime
from typing import Generic, TypeVar

import dbt.exceptions
from dbt_rpc.contracts.rpc import (
    RemoteCompileResult,
    RemoteCompileResultMixin,
    RemoteRunResult,
    ResultTable,
)
from dbt.logger import GLOBAL_LOGGER as logger
from dbt.task.compile import CompileRunner
from dbt_rpc.rpc.error import dbt_error, RPCException, server_error


RPCSQLResult = TypeVar("RPCSQLResult", bound=RemoteCompileResultMixin)


def get_raw_and_compiled(compiled_node):
    compiled_prev = getattr(compiled_node, 'compiled_sql', None)
    compiled_new = getattr(compiled_node, 'compiled_code', None)
    compiled_sql = compiled_new or compiled_prev
    raw_prev = getattr(compiled_node, 'raw_sql', None)
    raw_new = getattr(compiled_node, 'raw_code', None)
    raw_sql = raw_new or raw_prev
    return raw_sql, compiled_sql


class GenericRPCRunner(CompileRunner, Generic[RPCSQLResult]):
    def __init__(self, config, adapter, node, node_index, num_nodes):
        CompileRunner.__init__(self, config, adapter, node, node_index, num_nodes)

    def handle_exception(self, e, ctx):
        logger.debug("Got an exception: {}".format(e), exc_info=True)
        if isinstance(e, dbt.exceptions.Exception):
            if isinstance(e, dbt.exceptions.RuntimeException):
                e.add_node(ctx.node)
            return dbt_error(e)
        elif isinstance(e, RPCException):
            return e
        else:
            return server_error(e)

    def before_execute(self):
        pass

    def after_execute(self, result):
        pass

    def compile(self, manifest):
        if not self.node.config.enabled:
            raise dbt.exceptions.raise_compiler_error(
                "Trying to compile a node that is disabled"
            )
        compiler = self.adapter.get_compiler()
        return compiler.compile_node(self.node, manifest, {}, write=False)

    @abstractmethod
    def execute(self, compiled_node, manifest) -> RPCSQLResult:
        pass

    @abstractmethod
    def from_run_result(self, result, start_time, timing_info) -> RPCSQLResult:
        pass

    def error_result(self, node, error, start_time, timing_info):
        raise error

    def ephemeral_result(self, node, start_time, timing_info):
        raise dbt.exceptions.NotImplementedException(
            "cannot execute ephemeral nodes remotely!"
        )


class RPCCompileRunner(GenericRPCRunner[RemoteCompileResult]):
    def execute(self, compiled_node, manifest) -> RemoteCompileResult:
        raw_sql, compiled_sql = get_raw_and_compiled(compiled_node)
        return RemoteCompileResult(
            raw_sql=raw_sql,
            compiled_sql=compiled_sql,
            node=compiled_node,
            timing=[],  # this will get added later
            logs=[],
            generated_at=datetime.utcnow(),
        )

    def from_run_result(
            self,
            result: RemoteRunResult,
            start_time,
            timing_info
    ) -> RemoteCompileResult:

        return RemoteCompileResult(
            raw_sql=result.raw_sql,
            compiled_sql=result.compiled_sql,
            node=result.node,
            timing=timing_info,
            logs=[],
            generated_at=datetime.utcnow(),
        )


class RPCExecuteRunner(GenericRPCRunner[RemoteRunResult]):
    def execute(self, compiled_node, manifest) -> RemoteRunResult:
        raw_sql, compiled_sql = get_raw_and_compiled(compiled_node)
        _, execute_result = self.adapter.execute(compiled_sql, fetch=True)

        table = ResultTable(
            column_names=list(execute_result.column_names),
            rows=[list(row) for row in execute_result],
        )

        return RemoteRunResult(
            raw_sql=raw_sql,
            compiled_sql=compiled_sql,
            node=compiled_node,
            table=table,
            timing=[],
            logs=[],
            generated_at=datetime.utcnow(),
        )

    def from_run_result(self, result: RemoteRunResult, start_time, timing_info) -> RemoteRunResult:
        return RemoteRunResult(
            raw_sql=result.raw_sql,
            compiled_sql=result.compiled_sql,
            node=result.node,
            table=result.table,
            timing=timing_info,
            logs=[],
            generated_at=datetime.utcnow(),
        )



#!/usr/bin/env python
import os
from setuptools import setup, find_namespace_packages


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


package_name = "dbt-rpc"
package_version = "0.2.1"
description = """ A JSON RPC server that provides an interface to programmically interact with dbt projects. """


setup(
    name=package_name,
    version=package_version,
    description=description,
    long_description=description,
    author="dbt Labs",
    author_email="info@dbtlabs.com",
    url="https://github.com/dbt-labs/dbt-rpc",
    packages=find_namespace_packages(include=['dbt_rpc', 'dbt_rpc.*']),
    include_package_data=True,
    test_suite='tests',
    entry_points={
        'console_scripts': [
            'dbt-rpc = dbt_rpc.__main__:main',
        ],
    },
    install_requires=[
        'json-rpc>=1.12,<2',
        'dbt-core>=1.3.0rc1'
    ],
    zip_safe=False,
    classifiers=[
        'Development Status :: 5 - Production/Stable',

        'License :: OSI Approved :: Apache Software License',

        'Operating System :: Microsoft :: Windows',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX :: Linux',

        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    python_requires=">=3.6.3",
)
